/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License
 */

package android.server.cts;

import static android.server.cts.ActivityAndWindowManagersState.DEFAULT_DISPLAY_ID;
import static android.server.cts.ActivityManagerState.STATE_RESUMED;

/**
 * Build: mmma -j32 cts/hostsidetests/services
 * Run: cts/hostsidetests/services/activityandwindowmanager/util/run-test android.server.cts.ActivityManagerAssistantStackTests
 */
public class ActivityManagerAssistantStackTests extends ActivityManagerTestBase {

    private static final String VOICE_INTERACTION_SERVICE = "AssistantVoiceInteractionService";

    private static final String TEST_ACTIVITY = "TestActivity";
    private static final String ANIMATION_TEST_ACTIVITY = "AnimationTestActivity";
    private static final String DOCKED_ACTIVITY = "DockedActivity";
    private static final String ASSISTANT_ACTIVITY = "AssistantActivity";
    private static final String LAUNCH_ASSISTANT_ACTIVITY_FROM_SESSION =
            "LaunchAssistantActivityFromSession";
    private static final String LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK =
            "LaunchAssistantActivityIntoAssistantStack";
    private static final String PIP_ACTIVITY = "PipActivity";

    private static final String EXTRA_ENTER_PIP = "enter_pip";
    private static final String EXTRA_LAUNCH_NEW_TASK = "launch_new_task";
    private static final String EXTRA_ASSISTANT_DISPLAY_ID = "assistant_display_id";
    private static final String EXTRA_FINISH_SELF = "finish_self";
    public static final String EXTRA_IS_TRANSLUCENT = "is_translucent";

    private static final String TEST_ACTIVITY_ACTION_FINISH_SELF =
            "android.server.cts.TestActivity.finish_self";

    private static int mAssistantDisplayId;

    public void setUp() throws Exception {
        super.setUp();
        enableAssistant();
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK);
        mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);
        ActivityManagerState.ActivityStack assistantStack
                = mAmWmState.getAmState().getStackById(ASSISTANT_STACK_ID);
        mAssistantDisplayId = assistantStack.mDisplayId;
        removeStacks(ASSISTANT_STACK_ID);
        disableAssistant();
    }

    public void testLaunchingAssistantActivityIntoAssistantStack() throws Exception {
        // Enable the assistant and launch an assistant activity
        enableAssistant();
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_FROM_SESSION);
        mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);

        // Ensure that the activity launched in the fullscreen assistant stack
        assertAssistantStackExists();
        assertTrue("Expected assistant stack to be fullscreen",
                mAmWmState.getAmState().getStackById(ASSISTANT_STACK_ID).isFullscreen());

        disableAssistant();
    }

    public void testAssistantStackZOrder() throws Exception {
        if (!supportsPip() || !supportsSplitScreenMultiWindow()
                || !assistantRunsOnPrimaryDisplay()) return;
        // Launch a pinned stack task
        launchActivity(PIP_ACTIVITY, EXTRA_ENTER_PIP, "true");
        mAmWmState.waitForValidState(mDevice, PIP_ACTIVITY, PINNED_STACK_ID);
        mAmWmState.assertContainsStack("Must contain pinned stack.", PINNED_STACK_ID);

        // Dock a task
        launchActivity(TEST_ACTIVITY);
        launchActivityInDockStack(DOCKED_ACTIVITY);
        mAmWmState.assertHasFullscreenTypeStackOnDisplay("Must contain fullscreen stack.",
                mAssistantDisplayId);
        mAmWmState.assertContainsStack("Must contain docked stack.", DOCKED_STACK_ID);

        // Enable the assistant and launch an assistant activity, ensure it is on top
        enableAssistant();
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_FROM_SESSION);
        mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);
        assertAssistantStackExists();

        mAmWmState.assertFrontStack("Pinned stack should be on top.", PINNED_STACK_ID);
        mAmWmState.assertFocusedStack("Assistant stack should be focused.", ASSISTANT_STACK_ID);

        disableAssistant();
    }

    public void testAssistantStackLaunchNewTask() throws Exception {
        enableAssistant();
        assertAssistantStackCanLaunchAndReturnFromNewTask();
        disableAssistant();
    }

    public void testAssistantStackLaunchNewTaskWithDockedStack() throws Exception {
        if (!supportsSplitScreenMultiWindow() || !assistantRunsOnPrimaryDisplay()) return;
        // Dock a task
        launchActivity(TEST_ACTIVITY);
        launchActivityInDockStack(DOCKED_ACTIVITY);
        mAmWmState.assertHasFullscreenTypeStackOnDisplay("Must contain fullscreen stack.",
                mAssistantDisplayId);
        mAmWmState.assertContainsStack("Must contain docked stack.", DOCKED_STACK_ID);

        enableAssistant();
        assertAssistantStackCanLaunchAndReturnFromNewTask();
        disableAssistant();
    }

    private void assertAssistantStackCanLaunchAndReturnFromNewTask() throws Exception {
        // Enable the assistant and launch an assistant activity which will launch a new task
        enableAssistant();
        launchActivityOnDisplay(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK, mAssistantDisplayId,
                EXTRA_LAUNCH_NEW_TASK, TEST_ACTIVITY,
                EXTRA_ASSISTANT_DISPLAY_ID, Integer.toString(mAssistantDisplayId));
        disableAssistant();

        // Ensure that the fullscreen stack is on top and the test activity is now visible
        mAmWmState.waitForValidState(mDevice, TEST_ACTIVITY, FULLSCREEN_WORKSPACE_STACK_ID);
        mAmWmState.assertFocusedActivity("TestActivity should be resumed", TEST_ACTIVITY);
        mAmWmState.assertActivityInFullscreenTypeStack(
                "TestActivity should go on a suitable fullscreen-type stack.",
                TEST_ACTIVITY);

        // Now, tell it to finish itself and ensure that the assistant stack is brought back forward
        executeShellCommand("am broadcast -a " + TEST_ACTIVITY_ACTION_FINISH_SELF);
        mAmWmState.waitForFocusedStack(mDevice, ASSISTANT_STACK_ID);
        mAmWmState.assertFrontStack("Assistant stack should be on top.", ASSISTANT_STACK_ID);
        mAmWmState.assertFocusedStack("Assistant stack should be focused.", ASSISTANT_STACK_ID);
    }

    public void testAssistantStackFinishToPreviousApp() throws Exception {
        // Launch an assistant activity on top of an existing fullscreen activity, and ensure that
        // the fullscreen activity is still visible and on top after the assistant activity finishes
        launchActivityOnDisplay(TEST_ACTIVITY, mAssistantDisplayId);
        enableAssistant();
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK,
                EXTRA_FINISH_SELF, "true");
        disableAssistant();
        mAmWmState.waitForValidState(mDevice, TEST_ACTIVITY, FULLSCREEN_WORKSPACE_STACK_ID);
        mAmWmState.waitForActivityState(mDevice, TEST_ACTIVITY, STATE_RESUMED);
        mAmWmState.assertFocusedActivity("TestActivity should be resumed", TEST_ACTIVITY);
        mAmWmState.assertActivityInFullscreenTypeStack(
                "TestActivity should go on a suitable fullscreen-type stack.",
                TEST_ACTIVITY);
    }

    public void testDisallowEnterPiPFromAssistantStack() throws Exception {
        enableAssistant();
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK,
                EXTRA_ENTER_PIP, "true");
        disableAssistant();
        mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);
        mAmWmState.assertDoesNotContainStack("Must not contain pinned stack.", PINNED_STACK_ID);
    }

    public void testTranslucentAssistantActivityStackVisibility() throws Exception {
        enableAssistant();
        // Go home, launch the assistant and check to see that home is visible
        removeStacks(FULLSCREEN_WORKSPACE_STACK_ID);
        launchHomeActivity();
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK,
                EXTRA_IS_TRANSLUCENT, String.valueOf(true));
        mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);
        assertAssistantStackExists();
        mAmWmState.assertHomeActivityVisible(true);

        // Launch a fullscreen app and then launch the assistant and check to see that it is
        // also visible
        removeStacks(ASSISTANT_STACK_ID);
        launchActivityOnDisplay(TEST_ACTIVITY, mAssistantDisplayId);
        launchActivity(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK,
                EXTRA_IS_TRANSLUCENT, String.valueOf(true));
        mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);
        assertAssistantStackExists();
        mAmWmState.assertVisibility(TEST_ACTIVITY, true);

        // Launch a fullscreen and docked app and then launch the assistant and check to see that it
        // is also visible.
        if (supportsSplitScreenMultiWindow() && assistantRunsOnPrimaryDisplay()) {
            removeStacks(ASSISTANT_STACK_ID);
            launchActivityInDockStack(DOCKED_ACTIVITY);
            launchActivity(TEST_ACTIVITY);
            mAmWmState.assertContainsStack("Must contain docked stack.", DOCKED_STACK_ID);
            launchActivity(LAUNCH_ASSISTANT_ACTIVITY_INTO_STACK,
                    EXTRA_IS_TRANSLUCENT, String.valueOf(true));
            mAmWmState.waitForValidState(mDevice, ASSISTANT_ACTIVITY, ASSISTANT_STACK_ID);
            assertAssistantStackExists();
            mAmWmState.assertVisibility(DOCKED_ACTIVITY, true);
            mAmWmState.assertVisibility(TEST_ACTIVITY, true);
        }
        disableAssistant();
    }

    public void testLaunchIntoSameTask() throws Exception {
        enableAssistant();

        // Launch the assistant
        launchActivityOnDisplay(LAUNCH_ASSISTANT_ACTIVITY_FROM_SESSION, mAssistantDisplayId);
        assertAssistantStackExists();
        mAmWmState.assertVisibility(ASSISTANT_ACTIVITY, true);
        mAmWmState.assertFocusedStack("Expected assistant stack focused", ASSISTANT_STACK_ID);
        assertEquals(1, mAmWmState.getAmState().getStackById(ASSISTANT_STACK_ID).getTasks().size());
        final int taskId = mAmWmState.getAmState().getTaskByActivityName(ASSISTANT_ACTIVITY)
                .mTaskId;

        // Launch a new fullscreen activity
        // Using Animation Test Activity because it is opaque on all devices.
        launchActivityOnDisplay(ANIMATION_TEST_ACTIVITY, mAssistantDisplayId);
        mAmWmState.assertVisibility(ASSISTANT_ACTIVITY, false);

        // Launch the assistant again and ensure that it goes into the same task
        launchActivityOnDisplay(LAUNCH_ASSISTANT_ACTIVITY_FROM_SESSION, mAssistantDisplayId);
        assertAssistantStackExists();
        mAmWmState.assertVisibility(ASSISTANT_ACTIVITY, true);
        mAmWmState.assertFocusedStack("Expected assistant stack focused", ASSISTANT_STACK_ID);
        assertEquals(1, mAmWmState.getAmState().getStackById(ASSISTANT_STACK_ID).getTasks().size());
        assertEquals(taskId,
                mAmWmState.getAmState().getTaskByActivityName(ASSISTANT_ACTIVITY).mTaskId);

        disableAssistant();
    }

    // Any 2D Activity in VR mode is run on a special VR virtual display, so check if the Assistant
    // is going to run on the same display as other tasks.
    protected boolean assistantRunsOnPrimaryDisplay() {
        return mAssistantDisplayId == DEFAULT_DISPLAY_ID;
    }

    /**
     * Asserts that the assistant stack exists.
     */
    private void assertAssistantStackExists() throws Exception {
        mAmWmState.assertContainsStack("Must contain assistant stack.", ASSISTANT_STACK_ID);
    }

    /**
     * Asserts that the assistant stack does not exist.
     */
    private void assertAssistantStackDoesNotExist() throws Exception {
        mAmWmState.assertDoesNotContainStack("Must not contain assistant stack.",
                ASSISTANT_STACK_ID);
    }

    /**
     * Sets the system voice interaction service.
     */
    private void enableAssistant() throws Exception {
        executeShellCommand("settings put secure voice_interaction_service " +
                getActivityComponentName(VOICE_INTERACTION_SERVICE));
    }

    /**
     * Resets the system voice interaction service.
     */
    private void disableAssistant() throws Exception {
        executeShellCommand("settings delete secure voice_interaction_service " +
                getActivityComponentName(VOICE_INTERACTION_SERVICE));
    }
}
